package org.openscm.kundo.plugins

/*
 * Copyright (C) 2008 The Ultimate People Company Ltd ("UPCO").
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.openscm.kundo.plugins.context.BuildContext

/**
 * Cobertura Delegate
 * @author <a href="mailto:ramesh.ramakrishnan@upco.co.uk">Ramesh Ramakrishnan</a>
 * @version 1.0.0
 * 
 * Description: Cobertura plugin performs code coverage analysis for a given project
 * 
 * <p>
 * Code Coverage using cobertura involves 4 stages. 1. Code is instrumented 2.Junit testing is done.
 * 3. Adherence to Quality metrics is performed based on the data provided 4. Report gets generated.
 * For Adherence to quality, different metrics can be provided based on the package type. 
 * If project level has a packagelinerate of 85%, for specific packages it can be provided as 95%.
 * This is achieved by providing packageName, branchrate and linerate in the properties in the following format.
 * [cobertura].[check].[index].[packageName], [cobertura].[check].[index].[linerate], [cobertura].[check].[index].[branchrate]
 * </p>
 */
/* 
 * @ant-target doCobertura
 */
class CoberturaDelegate extends AbstractPluginTargetDelegate {

	protected Map projectProperties
	protected List srcProjectProperties
	protected List testProjectProperties
	protected List checkProjectProperties
	protected Map indexedProps
	/*
	 * Property to determine if code coverage to be performed using Cobertura plugin
	 * @ant-property-name cobertura.enabled
	 */
	 protected String enabled
	/*
	 * Directory containing the generated coverage reports in html format
	 * @ant-property-name cobertura.html.report.dir
	 */
	 protected String htmlReportDir
	/*
	 * Directory containing generated coverage file in xml format 
	 * @ant-property-name cobertura.xml.report.dir
	 */
	 protected String xmlReportDir
	/*
	 * Directory containing the instrumented classes
	 * @ant-property-name cobertura.instrumented.dir
	 */
	 protected String instrumentedDir
	/*
	 * Directory containing cobertura data file
	 * @ant-property-name cobertura.datafile.dir
	 */
	 protected String datafileDir
	/*
	 * name of data file
	 * @ant-property-name cobertura.datafile
	 */
	 protected String datafile
	/*
	 * Specify the minimum acceptable branch coverage report required for each classes
	 * @ant-property-name cobertura.branch.rate
	 */
	 protected String branchRate
	/*
	 * Specify the minimum acceptable line coverage report required for each classes
	 * @ant-property-name cobertura.line.rate
	 */
	 protected String lineRate
	/*
	 * This property when set to true will fail the build if none of the coverage metrics is met by the project
	 * @ant-property-name cobertura.halt.on.failure
	 */
	 protected String haltOnFailure
	/*
	 * Specify the minimum acceptable branch coverage report for each package
	 * @ant-property-name cobertura.package.branch.rate
	 */
	 protected String packageBranchRate
	/*
	 * Specify the minimum acceptable line coverage report required for each package
	 * @ant-property-name cobertura.package.line.rate
	 */
	 protected String packageLineRate
	/*
	 * Specify the minimum acceptable branch coverage report for the whole project
	 * @ant-property-name cobertura.total.branch.rate
	 */
	 protected String totalBranchRate
	/*
	 * Specify the minimum acceptable line coverage report for the whole project
	 * @ant-property-name cobertura.total.line.rate
	 */
	 protected String totalLineRate
	/*
	 * Specify the amount of memory to be used by JVM when processing large number of classes. 
	 * This property is used by both cobertura-instrument and cobertura-report tasks.
	 * @ant-property-name cobertura.max.memory
	 */
	 protected String maxMemory
	/*
	 * Print one-line statistic for each test case
	 * @ant-property-name cobertura.junit.print.summary
	 */
	 protected String junitPrintSummary
	/*
	 * Runs the junit test in separate VM
	 * @ant-property-name cobertura.junit.fork
	 */
	 protected String junitFork
	/*
	 * Controls how many VMs get created
	 * @ant-property-name cobertura.junit.fork.mode
	 */
	 protected String junitForkMode
	/*
	 * Type to generate Junit result
	 * @ant-property-name cobertura.junit.formatter
	 */
	 protected String junitFormatter
	/*
	 * File pattern for Including files for instrument task
	 * @ant-property-name cobertura.instrument.include
	 */
	 protected String instrumentInclude
	/*
	 * File pattern for Excluding files for instrument task
	 * @ant-property-name cobertura.instrument.exclude
	 */
	 protected String instrumentExclude
	/*
	 * File pattern for Including files for junit task
	 * @ant-property-name cobertura.junit.include
	 */
	 protected String junitInclude
	/*
	 * File pattern for Excluding files for junit task
	 * @ant-property-name cobertura.junit.exclude
	 */
	 protected String junitExclude
	/*
	 * File pattern for Including files for cobertura report task
	 * @ant-property-name cobertura.report.include
	 */
	 protected String reportInclude
	/*
	 * File pattern for Excluding files for cobertura report task
	 * @ant-property-name cobertura.report.exclude
	 */
	 protected String reportExclude
	 
	 /**
	 * Constructor sets ant and buildContext instances in super class
	 * @param ant			AntBuilder instance
	 * @param buildContext	BuildContext instance
	 */
	 CoberturaDelegate( AntBuilder ant, BuildContext buildContext ){
		super( ant, buildContext)
	}
	
	/**
	 * Setter method for member variable enabled
	 * @param enabled	
	 */
	void setEnabled( String enabled ){
		this.enabled = enabled
	}
	/**
	 * Setter method for member variable htmlReportDir
	 * @param htmlReportDir	
	 */
	void setHtmlReportDir( String htmlReportDir ){
		this.htmlReportDir = htmlReportDir
	}
	/**
	 * Setter method for member variable xmlReportDir
	 * @param xmlReportDir	
	 */
	void setXmlReportDir( String xmlReportDir ){
		this.xmlReportDir = xmlReportDir
	}
	/**
	 * Setter method for member variable instrumentedDir
	 * @param instrumentedDir	
	 */
	void setInstrumentedDir( String instrumentedDir ){
		this.instrumentedDir = instrumentedDir
	}
	/**
	 * Setter method for member variable datafileDir
	 * @param datafileDir	
	 */
	void setDatafileDir( String datafileDir ){
		this.datafileDir = datafileDir
	}
	/**
	 * Setter method for member variable datafile
	 * @param datafile	
	 */
	void setDatafile( String datafile ){
		this.datafile = datafile
	}
	/**
	 * Setter method for member variable branchRate
	 * @param branchRate	
	 */
	void setBranchRate( String branchRate ){
		this.branchRate = branchRate
	}
	/**
	 * Setter method for member variable lineRate
	 * @param lineRate	
	 */
	void setLineRate( String lineRate ){
		this.lineRate = lineRate
	}
	/**
	 * Setter method for member variable haltOnFailure
	 * @param haltOnFailure	
	 */
	void setHaltOnFailure( String haltOnFailure ){
		this.haltOnFailure = haltOnFailure
	}
	/**
	 * Setter method for member variable packageBranchRate
	 * @param packageBranchRate	
	 */
	void setPackageBranchRate( String packageBranchRate ){
		this.packageBranchRate = packageBranchRate
	}
	/**
	 * Setter method for member variable packageLineRate
	 * @param packageLineRate	
	 */
	void setPackageLineRate( String packageLineRate ){
		this.packageLineRate = packageLineRate
	}
	/**
	 * Setter method for member variable totalBranchRate
	 * @param totalBranchRate	
	 */
	void setTotalBranchRate( String totalBranchRate ){
		this.totalBranchRate = totalBranchRate
	}
	/**
	 * Setter method for member variable totalLineRate
	 * @param totalLineRate	
	 */
	void setTotalLineRate( String totalLineRate ){
		this.totalLineRate = totalLineRate
	}
	/**
	 * Setter method for member variable maxMemory
	 * @param maxMemory	
	 */
	void setMaxMemory( String maxMemory ){
		this.maxMemory = maxMemory
	}
	/**
	 * Setter method for member variable junitPrintSummary
	 * @param junitPrintSummary	
	 */
	void setJunitPrintSummary( String junitPrintSummary ){
		this.junitPrintSummary = junitPrintSummary
	}
	/**
	 * Setter method for member variable junitFork
	 * @param junitFork	
	 */
	void setJunitFork( String junitFork ){
		this.junitFork = junitFork
	}
	/**
	 * Setter method for member variable junitForkMode
	 * @param junitForkMode	
	 */
	void setJunitForkMode( String junitForkMode ){
		this.junitForkMode = junitForkMode
	}
	/**
	 * Setter method for member variable junitFormatter
	 * @param junitFormatter	
	 */
	void setJunitFormatter( String junitFormatter ){
		this.junitFormatter = junitFormatter
	}
	/**
	 * Setter method for member variable instrumentInclude
	 * @param instrumentInclude	
	 */
	void setInstrumentInclude( String instrumentInclude ){
		this.instrumentInclude = instrumentInclude
	}
	/**
	 * Setter method for member variable instrumentExclude
	 * @param instrumentExclude	
	 */
	void setInstrumentExclude( String instrumentExclude ){
		this.instrumentExclude = instrumentExclude
	}
	/**
	 * Setter method for member variable junitInclude
	 * @param junitInclude	
	 */
	void setJunitInclude( String junitInclude ){
		this.junitInclude = junitInclude
	}
	/**
	 * Setter method for member variable junitExclude
	 * @param junitExclude	
	 */
	void setJunitExclude( String junitExclude ){
		this.junitExclude = junitExclude
	}
	/**
	 * Setter method for member variable reportInclude
	 * @param reportInclude	
	 */
	void setReportInclude( String reportInclude ){
		this.reportInclude = reportInclude
	}
	/**
	 * Setter method for member variable reportExclude
	 * @param reportExclude	
	 */
	void setReportExclude( String reportExclude ){
		this.reportExclude = reportExclude
	}
	/**
	 * Setter method for member variable indexedProps
	 * @param indexedProps	Map of indexed properties to be injected
	 */		
	void setIndexedProps( Map indexedProps ){
		this.indexedProps = indexedProps
	}

	/**
	 * Implementation of abstract method from AbstractPluginTargetDelegate
	 * to hold assertions that all global and local properties are available to the plugin.
	 * This method is automatically called prior to execution.
	 */
	void doCheck(){
		assert enabled != null
		assert (enabled.equalsIgnoreCase("true") || enabled.equalsIgnoreCase("false"))
		assert htmlReportDir != null
		assert xmlReportDir != null
		assert instrumentedDir != null
		assert datafileDir != null
		assert datafile != null
		assert branchRate != null
		assert lineRate != null
		assert haltOnFailure != null
		assert packageBranchRate != null
		assert packageLineRate != null
		assert totalBranchRate != null
		assert totalLineRate != null
		assert maxMemory != null
		assert junitFork != null
		assert junitForkMode != null
		assert junitFormatter != null
		assert instrumentInclude != null
		assert instrumentExclude != null
		assert junitInclude != null
		assert junitExclude != null
		assert reportInclude != null
		assert reportExclude != null
		assert indexedProps != null
		assert buildContext.get( "project.properties" ) != null
	}
	
	/**
	 * Implementation of abstract method from AbstractPluginTargetDelegate
	 * this method is automatically called.
	 */	
	void doExecute(){		
		if (enabled.equalsIgnoreCase("true"))
			cobertura()
	}
	
	/**
	 * Private method for performing Code coverage using Cobertura
	 * This is done in 4 phases, Instrumentation code coverage, run test using Junit, perform adherence check and
	 * generate Coverage report
	 */
	private void cobertura() {
		 /* Task definition */
		 ant.'taskdef'( resource:"tasks.properties", classpathref:"kundo-cobertura-plugin.dependency.path")
		 /* Initialise required props */
		 projectProperties = buildContext.get( "project.properties" )
		 srcProjectProperties = projectProperties.get( "src" )
		 testProjectProperties = projectProperties.get( "test" )
		 checkProjectProperties = indexedProps.get( "check" )
		 defineClasspath( projectProperties )
		 /* Instrument the Classes */
		 performInstrument()
		 /* Run JUnit on test code */
		 performTest()
		 /* Check quality adherence */
		 performAdherenceCheck()
		 /* Generate code coverage Report*/
		 generateReport()
	 }
	
	/** 
	 * Defines class path variables for use during testing.
	 * @param projectProperties	 Map of project properties
	 */
	def defineClasspath( projectProperties )
	{
		ant.path( id:"junit.classpath" ) {
			ant.path( refid:"compile.classpath" )
			ant.path( refid:"test.compile.classpath" )
			projectProperties.each{ mapKey, propertyObjectList ->
			/* Check to see if there is a specified output directory */
				if( mapKey.equals( "test" ) )	{
					propertyObjectList.each{ propertyObject ->
						ant.pathelement( path:propertyObject.output )
					}					
				}				
			}
	
		}
	}
	
	/**
	 *  Private method for instrumenting the code 
	 */
	private void performInstrument() {
		/* Clean up for a new start */
		ant.delete( file : datafile)
		ant.delete( dir : instrumentedDir)
		ant.mkdir( dir : datafileDir)
		def buildDir = ant.project.getProperty( "global.build.dir" )
		def instrumentOptions = ["todir" : instrumentedDir,
		                          "datafile" : datafile]
		if (!maxMemory.equals("")) 
			instrumentOptions.put( "maxmemory" , maxMemory )
		
		ant.'cobertura-instrument'( instrumentOptions ) {
			/* Ignore source code having reference to log4j*/
			ant.ignore(regex:"org.apache.log4j.*")
			ant.fileset( dir:buildDir) {
				ant.include ( name : instrumentInclude )
				ant.exclude ( name : instrumentExclude )
			}
		}
	}
	
	/**
	 *  Private method for running JUnit test 
	 */
	private void performTest() {
		ant.mkdir ( dir : xmlReportDir )
		ant.mkdir ( dir : htmlReportDir )
		def junitOptions = [ "printsummary":junitPrintSummary,
		                     "fork":junitFork,
		                     "forkmode":junitForkMode]
		def sysKeyOptions = [ "key":"net.sourceforge.cobertura.datafile",
		                      "file":datafile]
		ant.junit ( junitOptions ) {
			ant.sysproperty( sysKeyOptions )
			/* Important to have the instrumented class to be in classpath before the uninstrumented class */
			ant.classpath( location:instrumentedDir )
			ant.classpath( refid:"junit.classpath" )
			ant.classpath( refid:"kundo-cobertura-plugin.dependency.path" )
			ant.formatter( type:junitFormatter )
			ant.'batchtest' ( todir:xmlReportDir ) {
				testProjectProperties.each{	testProjectProperty ->
					ant.fileset ( dir : testProjectProperty.dir) {
						ant.include( name : junitInclude )
				 		ant.exclude( name : junitExclude )
				 	}
				 }
			}
		}
	}
	
	/**
	 *  Private method to check if Project adheres to defined quality metrics 
	 */
	private void performAdherenceCheck() {
		Expando packageProp
		def checkOptions = [ "datafile" : datafile,
		                     "haltonfailure" : haltOnFailure,
		                     "branchrate" : branchRate,
		                     "linerate" : lineRate,
		                     "totalbranchrate" : totalBranchRate,
		                     "totallinerate" : totalLineRate,
		                     "packagebranchrate" : packageBranchRate,
		                     "packagelinerate" : packageLineRate]
		ant.'cobertura-check'( checkOptions ) {
			checkProjectProperties.eachWithIndex { checkPackageProp, index ->
				def pkgOptions = ["pattern" : checkPackageProp.packageName,
				                  "linerate":checkPackageProp.linerate,
				                  "branchrate":checkPackageProp.branchrate]
				ant.regex( pkgOptions )
			}
		}
	}
	
	/** 
	 * Private method for generating Report 
	 */
	private void generateReport() {
		/* Generate XML file containing coverage data using "srcdir" attribute*/
		srcProjectProperties.each{	srcProjectProperty ->
			ant.'cobertura-report'(	srcdir : srcProjectProperty.dir,
		 							destdir : xmlReportDir,
		 							format : "xml",
		 							datafile : datafile)
		}
		/* Generate HTML Report */
		def reportOptions = [ "destdir":htmlReportDir,
		                      "datafile":datafile,
		                      "format":"html"]
		if (!maxMemory.equals(""))
			reportOptions.put( "maxmemory" , maxMemory )
		ant.'cobertura-report'( reportOptions ) {
			srcProjectProperties.each{	srcProjectProperty ->
				ant.fileset( dir : srcProjectProperty.dir) {
					ant.include ( name : reportInclude )
			 		ant.exclude ( name : reportExclude )
			 	}
			 }
		}
	}
	
}